# A Spritefile input filter for Sketch (http://sketch.sourceforge.net)
# Copyright (C) 2002-2003 by David Boddie
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Versions history:
# 
# 0.10 (Sat 16th March 2002)
# 
# First release version based on the drawinput filter.

###Sketch Config
#type = Import
#class_name = 'SpritefileLoader'
#rx_magic = '^....\\x10\\x00\\x00\\x00'
#tk_file_type = ('Spritefile', ('.ff9', ',ff9'))
format_name = 'Spritefile'
#standard_messages = 1
#unload = 1
###End

(''"Spritefile")

version = '0.10 (Sat 16th March 2002)'

import os, string

try:
    import cStringIO
    StringIO = cStringIO
except ImportError:
    import StringIO

from math import floor

from Sketch import Document, Layer, CreatePath, ContSmooth, \
     SolidPattern, EmptyPattern, LinearGradient, RadialGradient,\
     CreateRGBColor, CreateCMYKColor, MultiGradient, Trafo, Point, Polar, \
     StandardColors, GetFont, PathText, SimpleText, const, UnionRects, \
     Arrow, Translation, Trafo, GridLayer, Scale

from Sketch.load import GenericLoader, EmptyCompositeError

from Sketch.Graphics import pagelayout

from Sketch import SolidPattern, EmptyPattern

from PIL import Image

from Lib.Drawfile import spritefile

papersizes = [
        #    'A0', 'A1', 'A2',
        'A3', 'A4', 'A5', 'A6', 'A7',
        'letter', 'legal', 'executive',
        ]

orientations = ['portrait', 'landscape']

class SpritefileLoader(GenericLoader):
    """
    SpritefileLoader

    Read a Spritefile and store the elements in a Sketch document
    """

    format_name = format_name

    def __init__(self, file, filename, match):

        GenericLoader.__init__(self, file, filename, match)

        # GenericLoader puts file, filename and match into object attributes
        # with the same names

    def Load(self):

        # Page layout
        self.page_layout = None

        # Instantiate a spritefile object
        try:
            self.spritefile = spritefile.spritefile(self.file)
        
        except spritefile.spritefile_error:
        
            self.add_message(_('Failed to read Spritefile '+self.filename))
            self.debugout.write('Failed to read Spritefile '+self.filename+'\n')
        
        # Create the document (creates self.object).
        self.document()

        # Insert a layer
        self.layer('Layer 1')

        # Read the sprites.
        try:
            width, height = self.read_sprites()
        
        except spritefile.spritefile_error:
        
            self.add_message(_('Failed to read Spritefile '+self.filename))
            self.debugout.write('Failed to read Spritefile '+self.filename+'\n')

        self.end_all()

        # Page layout
        # Create a page which is large enough to contain the sprites.
        
        #self.object.load_SetLayout(
        #    pagelayout.PageLayout(width = width, height = height) )

        self.object.load_Completed()

        # Return the document.
        return self.object

    def read_sprites(self):
    
        # Create dictionaries containing the widths, heights, scale factors
        # and areas of each sprite.
        widths = {}
        heights = {}
        areas = {}
        xscales = {}
        yscales = {}
        
        # Create a list of names of sprites of descending area.
        order = []
        
        for name, sprite in self.spritefile.sprites.items():
        
            # Convert the sprite dimensions to page units:
            scale_x = (72.0 / self.spritefile.sprites[name]['dpi x'])
            scale_y = (72.0 / self.spritefile.sprites[name]['dpi y'])
            
            swidth = self.spritefile.sprites[name]['width'] * scale_x
            sheight = self.spritefile.sprites[name]['height'] * \
                      scale_y
            
            widths[name] = swidth
            heights[name] = sheight
            xscales[name] = scale_x
            yscales[name] = scale_y
            
            area = swidth * sheight
            areas[name] = area
            
            i = 0
            for test in order:
            
                if areas[test] < area:
                
                    # We will insert this name in the list here.
                    break
                
                i = i + 1
            
            # Insert the name into the area order list.
            order.insert(i, name)
        
        # The sprites still to be placed.
        remaining = self.spritefile.sprites.keys()
        
        # The area covered by sprites.
        used_area = [0.0, 0.0]
        
        # A list of free areas with this area.
        free = []
        
        # Traverse the list of sprites, inserting them in the page.
        while len(remaining) > 0:
        
            # If there are free areas available then try to fill them
            # with suitably sized sprites.
            
            if free != []:
            
                # Try to merge adjacent areas.
                new_free = []
                
                while free != []:
                
                    # Choose a free area.
                    free1 = free.pop()
                    
                    merged = 0
                    
                    for free2 in free:
                    
                        if free1[0] == free2[0] and \
                            free1[2] == free2[2]:
                        
                            # Same x position and width.
                            if free1[1] + free1[3] >= free2[1]:
                            
                                # The other area is above this one.
                                new_free.append(
                                    [ free1[0], free1[1],
                                      free1[2], free1[3] + free2[3] ] )
                                
                                # Remove the second area from the list and
                                # stop comparisons.
                                free.remove(free2)
                                merged = 1
                                break
                            
                        elif free1[1] == free2[1] and \
                            free1[3] == free2[3]:
                        
                            # Same y position and height.
                            if free1[0] + free1[2] >= free2[0]:
                            
                                # The other area is to the right of this one.
                                # The other area is above this one.
                                new_free.append(
                                    [ free1[0], free1[1],
                                      free1[2], free1[3] + free2[3] ] )
                                
                                # Remove the second area from the list and
                                # stop comparisons.
                                free.remove(free2)
                                merged = 1
                                break
                    
                    # If the area was not merged with another then add it
                    # to the new list.
                    if merged == 0:
                    
                        new_free.append(free1)
                    
                # Copy the new free list over the old one.
                free = new_free
                
                # Look in the list of remaining sprites for one which will
                # fit within each area.
                
                used = []
                
                for name in remaining:
                
                    # Retrieve the sprite dimensions in page units:
                    
                    filled = []
                    
                    for f in free:
                    
                        x, y, width, height = f[0], f[1], f[2], f[3]
                        
                        if widths[name] <= width and heights[name] <= height:
                        
                            # It fits: place the sprite in here.
                            self.place_image( name, x, y,
                                              xscales[name], yscales[name] )
                            
                            if widths[name] == width and \
                                heights[name] == height:
                            
                                # We have filled this area.
                                filled.append(f)
                            
                            elif widths[name] < width and \
                                heights[name] == height:
                            
                                # Make this area smaller horizontally.
                                f[0] = f[0] + widths[name]
                                f[2] = f[2] - widths[name]
                            
                            elif widths[name] == width and \
                                heights[name] < height:
                            
                                # Make this area smaller vertically.
                                f[1] = f[1] + heights[name]
                                f[3] = f[3] - heights[name]
                            
                            else:
                            
                                # Make this area smaller horizontally
                                # and create a new area.
                                free.append( [ f[0], f[1] + heights[name],
                                               widths[name],
                                               f[3] - heights[name] ] )
                                
                                f[0] = f[0] + widths[name]
                                f[2] = f[2] - widths[name]
                                
                            # This sprite has been used.
                            used.append(name)
                            
                            break
                    
                    # Remove any filled areas from the list of free areas.
                    for f in filled:
                    
                        free.remove(f)
                
                # Remove any used sprites from the list of remaining ones.
                for u in used:
                
                    remaining.remove(u)
            

            # Look at the general area covered by sprites.
            
            if used_area[0] < used_area[1] and remaining != []:
            
                # The sprites are taking up more vertical space than
                # horizontal.
                
                # Look for the tallest sprite which will fit alongside the
                # area. We want one which is comparable in height with
                # the current filled area, but with minimal width.
                
                found = remaining[0]
                
                # Calculate a value for the coefficient of suitability for
                # the first remaining sprite:
                best = abs(used_area[1] - heights[found]) * widths[found]
                
                for name in remaining[1:]:
                
                    # Determine the value of the coefficient for this
                    # sprite.
                    c = abs(used_area[1] - heights[name]) * widths[name]
                    
                    # If the value of the coefficient for this sprite
                    # is lower than that for the previous best sprite
                    # then make this sprite the best sprite.
                    if c < best:
                    
                        found = name
                        best = c
                    
                    elif c == best:
                    
                        # The coefficients are the same. Choose the
                        # taller sprite.
                        if heights[name] > heights[found]:
                        
                            found = name
                            best = c
                
                # Use the last sprite looked at.
                self.place_image( found, used_area[0], 0.0,
                                  xscales[found], yscales[found] )
                
                if heights[found] > used_area[1]:
                
                    # We have generated another free area.
                    #
                    #       used_area[0]
                    #           |
                    # +---------+----+ sheight
                    # |#########|    |
                    # |#########|    |
                    # +---------+    | used_area[1]
                    # |         |    |
                    # |         |    |
                    # |         |    |
                    # |         |    |
                    # +---------+----+ 0.0
                    #
                    # 0.0            used_area[0]+swidth
                    
                    free.append( [ 0.0, used_area[1],
                                   used_area[0],
                                   heights[found] - used_area[1] ] )
                
                elif heights[found] < used_area[1]:
                
                    # We have generated another free area.
                    #
                    #       used_area[0]
                    #           |
                    # +---------+----| used_area[1]
                    # |         |####|
                    # |         |----| sheight
                    # |         |    |
                    # |         |    |
                    # +---------+----+ 0.0
                    #
                    # 0.0            area[0]+swidth
                    
                    free.append( [ used_area[0], heights[found],
                                   widths[found],
                                   used_area[1] - heights[found] ] )
                
                # We have extended the used area.
                used_area[0] = used_area[0] + widths[found]
                used_area[1] = max(used_area[1], heights[found])
                
                # This sprite has been used.
                remaining.remove(found)
            
            elif used_area[1] < used_area[0] and remaining != []:
            
                # The sprites are taking up more horizontal space than
                # vertical.
                
                # Look for the widest sprite which will fit alongside the
                # area. We want one which is comparable in width, but
                # with minimal height.
                
                found = remaining[0]
                
                # Calculate a value for the coefficient of suitability for
                # the first remaining sprite:
                best = abs(used_area[1] - widths[found]) * heights[found]
                
                for name in remaining[1:]:
                
                    # Determine the value of the coefficient for this
                    # sprite.
                    c = abs(used_area[1] - widths[name]) * heights[name]
                    
                    # If the value of the coefficient for this sprite
                    # is lower than that for the previous best sprite
                    # then make this sprite the best sprite.
                    if c < best:
                    
                        found = name
                        best = c
                    
                    elif c == best:
                    
                        # The coefficients are the same. Choose the
                        # wider sprite.
                        if widths[name] > widths[found]:
                        
                            found = name
                            best = c
                
                # Use the last sprite looked at.
                self.place_image( found, 0.0, used_area[1],
                                  xscales[found], yscales[found] )
                
                if widths[found] > used_area[0]:
                
                    # We have generated another free area.
                    #
                    #       used_area[0]
                    #           |
                    # +---------+----+ used_area[1]+sheight
                    # |              |
                    # |              |
                    # +---------+----| used_area[1]
                    # |         |####|
                    # |         |####|
                    # |         |####|
                    # |         |####|
                    # +---------+----+ 0.0
                    #
                    # 0.0            swidth
                    
                    free.append( [ used_area[0], 0.0,
                                   widths[found] - used_area[0],
                                   used_area[1] ] )
                
                elif widths[found] < used_area[0]:
                
                    # We have generated another free area.
                    #
                    #      swidth
                    #        |
                    # +------+--+      used_area[1]+sheight
                    # |      |##|
                    # |      |##|
                    # +---------+      used_area[1]
                    # |         |
                    # |         |
                    # |         |
                    # |         |
                    # +---------+      0.0
                    #
                    # 0.0       used_area[0]
                    
                    free.append( [ widths[found], used_area[1],
                                   used_area[0] - widths[found],
                                   heights[found] ] )
                
                # We have extended the used area.
                used_area[0] = max(used_area[0], widths[found])
                used_area[1] = used_area[1] + heights[found]
                
                # This sprite has been used.
                remaining.remove(found)
                        
            elif remaining != []:
            
                # The area used is square.
                
                # Take the piece with the greatest area and place it above
                # the used area.
                
                for name in order:
                
                    if name in remaining:
                    
                        break
                
                # Place the sprite.
                self.place_image( name, 0.0, used_area[1],
                                  xscales[name], yscales[name] )
                
                if widths[name] > used_area[0] and used_area[1] != 0.0:
                
                    # We have generated another free area.
                    #
                    #       used_area[0]
                    #           |
                    # +---------+----+ used_area[1]+sheight
                    # |              |
                    # |              |
                    # +---------+----| used_area[1]
                    # |         |####|
                    # |         |####|
                    # |         |####|
                    # |         |####|
                    # +---------+----+ 0.0
                    #
                    # 0.0            swidth
                    
                    free.append( [ used_area[0], 0.0,
                                   widths[name] - used_area[0],
                                   used_area[1] ] )
                
                elif widths[name] < used_area[0]:
                
                    # We have generated another free area.
                    #
                    #      swidth
                    #        |
                    # +------+--+      used_area[1]+sheight
                    # |      |##|
                    # |      |##|
                    # +---------+      used_area[1]
                    # |         |
                    # |         |
                    # |         |
                    # |         |
                    # +---------+      0.0
                    #
                    # 0.0       used_area[0]
                    
                    free.append( [ widths[name], used_area[1],
                                   used_area[0] - widths[name],
                                   heights[name] ] )
                
                # We have extended the used area.
                used_area[0] = max(used_area[0], widths[name])
                used_area[1] = used_area[1] + heights[name]
                
                # This sprite has been used.
                remaining.remove(name)
        
        # Return the extent of the area used.
        return used_area[0], used_area[1]
    
    def place_image(self, name, x, y, scale_x, scale_y):
    
        transform = Trafo( scale_x, 0.0, 0.0, scale_y, x, y )
        
        # Create an Image object
        image = Image.fromstring(self.spritefile.sprites[name]['mode'],
                                 (self.spritefile.sprites[name]['width'],
                                  self.spritefile.sprites[name]['height']),
                                 self.spritefile.sprites[name]['image'])
        
        self.image(image, transform)
            
    def make_safe(self, s):

        new = ''
        for i in s:
            if ord(i) >= 32:
                new = new + i
        return new


    def relative(self, x, y):

#        return (x - self.drawfile.x1)/scale, (y - self.drawfile.y1)/scale
        return x/scale, y/scale
